import copy
from typing import List

import isle._urls as URL
import isle.objects._tmdb as _tmdb_obj
import isle.objects._others as other_objs

from .._config import tmdb_api_key


__all__ = ["Company"]


class Company(_tmdb_obj.TMDb):
    """Represents a company."""

    def _init(self):
        self.get_details()

    def _getdata(self, key):
        if key not in self.data:
            if key == "alternative_names":
                self.get_alternative_names()
            elif key == "images":
                self.get_images()
            else:
                self._init()
        return copy.deepcopy(self.data[key])

    @property
    def name(self):
        """Return the name of a company."""
        return self._getdata("name")

    @property
    def also_known_as(self):
        """Return alternative names of a company."""
        names = []
        for item in self._getdata("alternative_names")["results"]:
            names.append(item["name"])
        return names

    @property
    def description(self):
        """Return a company's description."""
        return self._getdata("description")

    @property
    def homepage(self):
        """Return a company's homepage."""
        return self._getdata("homepage")

    @property
    def country(self):
        """Return a company's origin country."""
        code = self._getdata("origin_country")
        if code:
            try:
                english_name = self._all_countries[code]
            except AttributeError:
                self._all_countries = self._get_all_countries()
                english_name = self._all_countries[code]
            return other_objs.Country(
                iso_3166_1=code, english_name=english_name
            )
        else:
            return code

    @property
    def parent_company(self):
        # TODO: definition
        return self._getdata("parent_company")

    @property
    def logos(self):
        """Return logo images that belong to a comapny. Each item
        is an instance of the `Image` class."""
        logos = []
        for item in self._getdata("images")["logos"]:
            logos.append(other_objs.Image(item, type_="logo"))
        return logos

    def _get_all_countries(self):
        data = self._request(
            URL.COUNTRIES_CONFIGURATION, **{"api_key": tmdb_api_key()}
        )
        countries = {}
        for item in data:
            countries[item["iso_3166_1"]] = item["english_name"]
        return countries

    def get_details(self, **params):
        """Get a companies details."""
        details = self._request(
            URL.COMPANY_DETAILS.format(company_id=self.tmdb_id), **params
        )
        self.data.update(details)
        return details

    def get_alternative_names(self, **params):
        """Get the alternative names of a company."""
        alternative_names = self._request(
            URL.COMPANY_ALTERNATIVE_NAMES.format(company_id=self.tmdb_id),
            **params
        )
        self.data.update({"alternative_names": alternative_names})
        return alternative_names

    def get_images(self, **params):
        """Get a companies logos."""
        images = self._request(
            URL.COMPANY_IMAGES.format(company_id=self.tmdb_id), **params
        )
        self.data.update({"images": images})
        return images



import os
from collections import defaultdict
from urllib.parse import urljoin

from . import _urls as URL

from ._config import tmdb_api_key
from ._requests import GET, GET_pages
from .objects import (
    Company,
    Movie,
    Country,
    Genre,
    Language,
    Person,
    Episode,
    Season,
    Show,
)


__all__ = [
    "search_movie",
    "search_show",
    "search_person",
    "search_company",
    "discover_movies",
    "discover_shows",
    "find",
    "get_movie_certifications",
    "get_show_certifications",
    "get_movie_genres",
    "get_show_genres",
    "get_image_configurations",
    "get_countries",
    "get_jobs",
    "get_languages",
    "get_primary_translations",
    "get_timezones",
]


def search_movie(query: str, **kwargs):
    """Search for movies.

    The `query` argument is a text query to search (required).

    The optional `year` argument specifies the release year of
    the movie.

    The optional `language` argument specifies a ISO 639-1 code
    to display translated data for the fields that support it.
    (Default: "en-US")

    The optional `include_adult` argument specifies whether
    to include adult (pornography) content in the results.
    (Default: False)

    The optional `region` argument specifies a ISO 3166-1 code
    to filter release dates. Must be uppercase.

    Returns a generator. Each item is a `Movie` object.
    """
    params = {"query": query, "api_key": tmdb_api_key(), **kwargs}
    for item in GET_pages(URL.SEARCH_MOVIE, params):
        yield Movie(item["id"], **item)


def search_show(query: str, **kwargs):
    """Search for TV shows.

    The `query` argument is a text query to search (required).

    The optional `language` argument specifies a ISO 639-1 code
    to display translated data for the fields that support it.
    (Default: "en-US")

    The optional `first_air_date_year` argument specifies
    the year when the show was first aired.

    Returns a generator. Each item is a `Show` object.
    """
    params = {"query": query, "api_key": tmdb_api_key(), **kwargs}
    for item in GET_pages(URL.SEARCH_SHOW, params):
        yield Show(item["id"], **item)


def search_person(query: str, **kwargs):
    """Search for people.

    The `query` argument is a text query to search (required).

    The optional `language` argument specifies a ISO 639-1 code
    to display translated data for the fields that support it.
    (Default: "en-US")

    The optional `include_adult` argument specifies whether
    to include adult (pornography) content in the results.
    (Default: False)

    The optional `region` argument specifies a ISO 3166-1 code
    to filter release dates. Must be uppercase.

    Returns a generator. Each item is a `Person` object.
    """
    params = {"query": query, "api_key": tmdb_api_key(), **kwargs}
    for item in GET_pages(URL.SEARCH_PERSON, params):
        yield Person(item["id"], **item)


def search_company(query: str, **kwargs):
    """Search for companies.

    The `query` argument is a text query to search (required).

    Returns a generator. Each item is a `Company` object.
    """
    params = {"query": query, "api_key": tmdb_api_key(), **kwargs}
    for item in GET_pages(URL.SEARCH_COMPANY, params):
        yield Company(item["id"], **item)


def discover_movies(options: dict):
    """Discover movies by different types of data like
    average rating, number of votes, genres and certifications.

    See available options:
    https://developers.themoviedb.org/3/discover/movie-discover

    Returns a generator. Each item is a `Movie` object.
    """
    params = {"api_key": tmdb_api_key(), **options}
    for item in GET_pages(URL.DISCOVER_MOVIES, params):
        yield Movie(item["id"], **item)


def discover_shows(options: dict):
    """Discover TV shows by different types of data like
    average rating, number of votes, genres, the network
    they aired on and air dates.

    See available options:
    https://developers.themoviedb.org/3/discover/tv-discover

    Returns a generator. Each item is a `Show` object.
    """
    params = {"api_key": tmdb_api_key(), **options}
    for item in GET_pages(URL.DISCOVER_SHOWS, params):
        yield Show(item["id"], **item)


def find(external_id: str, *, src: str, **options):
    """Search for objects by an external id.

    Allowed sources (`src`): `imdb_id`, `freebase_mid`,
    `freebase_id`, `tvdb_id`, `tvrage_id`

    See available options:
    https://developers.themoviedb.org/3/find/find-by-id
    """
    params = {"api_key": tmdb_api_key(), "external_source": src, **options}
    response = GET(URL.FIND.format(external_id=external_id), **params)
    acc = {}
    for key, results in response.items():
        if key == "movie_results":
            acc[key] = [Movie(x["id"], **x) for x in results]
        elif key == "person_results":
            acc[key] = [Person(x["id"], **x) for x in results]
        elif key == "tv_results":
            acc[key] = [Show(x["id"], **x) for x in results]
        elif key == "tv_episode_results":
            for x in results:
                n = x["episode_number"]
                sn = x["season_number"]
                show_id = x["show_id"]
                del x["episode_number"]
                del x["season_number"]
                del x["show_id"]
                acc.setdefault(key, []).append(
                    Episode(n, show_id=show_id, season_number=sn, **x)
                )
        elif key == "tv_season_results":
            for x in results:
                n = x["season_number"]
                show_id = x["show_id"]
                del x["season_number"]
                del x["show_id"]
                acc.setdefault(key, []).append(Season(n, show_id=show_id, **x))
        else:
            raise ValueError(f"Unkown type: {key}")
    return acc


def get_movie_certifications(country=None):
    """Get an up to date list of the officially supported
    movie certifications on TMDb."""
    res = GET(URL.MOVIE_CERTIFICATION, **{"api_key": tmdb_api_key()})[
        "certifications"
    ]
    return res[country] if country else res


def get_show_certifications(country=None):
    """Get an up to date list of the officially supported TV
    show certifications on TMDb."""
    res = GET(URL.SHOW_CERTIFICATION, **{"api_key": tmdb_api_key()})[
        "certifications"
    ]
    return res[country] if country else res


def get_movie_genres(objects=False):
    """Get the list of official genres for movies."""
    genres = GET(URL.MOVIE_GENRES, **{"api_key": tmdb_api_key()})["genres"]
    if objects:
        return [Genre(tmdb_id=g["id"], name=g["name"]) for g in genres]
    else:
        return genres


def get_show_genres(objects=False):
    """Get the list of official genres for TV shows."""
    genres = GET(URL.SHOW_GENRES, **{"api_key": tmdb_api_key()})["genres"]
    if objects:
        return [Genre(tmdb_id=g["id"], name=g["name"]) for g in genres]
    else:
        return genres


def get_image_configurations():
    """Get the data relevant to building image URLs as well
    as the change key map."""
    return GET(URL.IMAGE_CONFIGURATION, **{"api_key": tmdb_api_key()})


def get_countries(objects=False):
    """Get the list of countries (ISO 3166-1 tags) used
    throughout TMDb."""
    countries = GET(URL.COUNTRIES_CONFIGURATION, **{"api_key": tmdb_api_key()})
    if objects:
        return [Country(**c) for c in countries]
    else:
        return countries


def get_jobs():
    """Get a list of the jobs and departments we use on
    TMDb."""
    return GET(URL.JOBS_CONFIGURATION, **{"api_key": tmdb_api_key()})


def get_languages(objects=False):
    """Get the list of languages (ISO 639-1 tags) used
    throughout TMDb."""
    languages = GET(URL.LANGUAGES_CONFIGURATION, **{"api_key": tmdb_api_key()})
    if objects:
        return [
            Language(
                iso_639_1=l["iso_639_1"],
                english_name=l["english_name"],
                original_name=l["name"],
            )
            for l in languages
        ]
    else:
        return languages


def get_primary_translations():
    """Get a list of the officially supported translations
    on TMDb."""
    return GET(
        URL.PRIMARY_TRANSLATIONS_CONFIGURATION, **{"api_key": tmdb_api_key()}
    )


def get_timezones():
    """Get the list of timezones used throughout TMDb."""
    return GET(URL.TIMEZONES_CONFIGURATION, **{"api_key": tmdb_api_key()})



from typing import Iterator, List, NamedTuple, Optional, Union

import isle._urls as URL
import isle.objects._tmdb as _tmdb_obj
import isle.objects._movie as movie_obj
import isle.objects._person as person_obj
import isle.objects._show as show_obj
from .._config import tmdb_api_key
from .._requests import GET


__all__ = [
    "Keyword",
    "Genre",
    "Image",
    "Country",
    "Language",
    "Vote",
    "Video",
    "Credit",
]


class Keyword(_tmdb_obj.TMDb):
    """Represents a keyword."""

    def _init(self):
        self.get_details()

    @property
    def name(self):
        """Return a keyword's name"""
        return self._getdata("name")

    def get_details(self):
        """Get the primary information about a keyword."""
        details = self._request(
            URL.KEYWORD_DETAILS.format(keyword_id=self.tmdb_id)
        )
        self.data.update(details)
        return details

    def iter_movies(self, **params):
        """Get the movies that belong to a keyword."""
        results = self._iter_request(
            URL.KEYWORD_MOVIES.format(keyword_id=self.tmdb_id), **params
        )
        yield from map(lambda x: movie_obj.Movie(x["id"]), results)

    def __str__(self):
        return self._getdata("name")


class Genre(NamedTuple):
    """Represents a genre."""

    tmdb_id: str
    name: str

    def __str__(self):
        return self.name


class Image:
    """Represents an image."""

    def __init__(self, image: dict, *, type_: str):
        assert type_ in ["backdrop", "poster", "logo", "profile", "still"]
        self._type = type_
        self._configs_data = {}
        for key in image.keys() - {"vote_average", "vote_count"}:
            setattr(self, key, image[key])
        if {"vote_average", "vote_count"} <= image.keys():
            self.vote = Vote(
                average=image["vote_average"], count=image["vote_count"]
            )

    @property
    def _configs(self):
        if not self._configs_data:
            self._configs_data.update(self._get_image_configs())
        return self._configs_data

    @property
    def url(self):
        urls = {}
        base = self._configs["secure_base_url"]
        for size in self.sizes:
            urls[size] = f"{base}/{size}{self.file_path}"
        return urls

    @property
    def sizes(self):
        return self._configs[self._image_sizes_key]

    @property
    def _image_sizes_key(self):
        if self._type == "backdrop":
            return "backdrop_sizes"
        elif self._type == "poster":
            return "poster_sizes"
        elif self._type == "logo":
            return "logo_sizes"
        elif self._type == "profile":
            return "profile_sizes"
        elif self._type == "still":
            return "still_sizes"
        else:
            raise ValueError(f"Unknown image type: {self._type}")

    def _get_image_configs(self):
        res = GET(URL.IMAGE_CONFIGURATION, **{"api_key": tmdb_api_key()})[
            "images"
        ]
        return res

    def __repr__(self):
        return f"Image(height={self.height}, width={self.width}, _type={self._type})"


class Country(NamedTuple):
    """Represents a country."""

    iso_3166_1: str
    english_name: str

    def __str__(self):
        return self.english_name


class Language(NamedTuple):
    """Represents a language."""

    iso_639_1: str
    english_name: str
    original_name: str

    def __str__(self):
        return self.english_name


class Vote(NamedTuple):
    """Represents a vote."""

    average: float
    count: str

    def __str__(self):
        return self.average


class Video(NamedTuple):
    """Represents a video."""

    name: str
    type: str
    url: str

    def __str__(self):
        return self.url


class Credit(_tmdb_obj.TMDb):
    """Represents a credit."""

    def __init__(
        self,
        tmdb_id,
        *,
        person_data=None,
        media_data=None,
        character=None,
        **kwargs,
    ):
        self.data = {"credit_id": tmdb_id, **kwargs}
        self.tmdb_id = self.data["credit_id"]
        self._person_data = person_data
        self._media_data = media_data
        self._character = character
        self.n_requests = 0

    def _init(self):
        self.get_details()

    @property
    def type(self):
        """Return a credit type"""
        return self._getdata("credit_type")

    @property
    def media_type(self):
        """Return a media type"""
        return self._getdata("media_type")

    @property
    def department(self):
        return self._getdata("department")

    @property
    def job(self):
        return self._getdata("job")

    @property
    def character(self):
        if self.type == "crew":
            return None
        return self._character or self._getdata("media")["character"]

    @property
    def person(self):
        data = self._person_data or self._getdata("person")
        return person_obj.Person(data["id"], **data)

    @property
    def person_known_for(self):
        media = []
        for item in self._getdata("person")["known_for"]:
            Obj = (
                show_obj.Show
                if item["media_type"] == "tv"
                else movie_obj.Movie
            )
            media.append(Obj(item["id"], **item))
        return media

    @property
    def media(self):
        Obj = show_obj.Show if self.media_type == "tv" else movie_obj.Movie
        if self._media_data:
            data = self._media_data
        else:
            data = self._getdata("media")
            data.pop("seasons", None)
            data.pop("episodes", None)
        return Obj(data["id"], **data)

    def get_details(self):
        """Get a movie or TV credit details."""
        details = self._request(
            URL.CREDIT_DETAILS.format(credit_id=self.tmdb_id)
        )
        self.data.update(details)
        return details
